import sgmllib
import Acquisition

class MockResponse(object):
    
    def __init__(self):
        self.status = None
        self.headers = {}
        self.body = ''
        
    def setStatus(self, status, lock=False):
        self.status = status
        
    def setHeader(self, key, value):
        self.headers[key] = value
        
    def setBody(self, body, lock=False):
        self.body = body
        

class MockRequest(object):
    
    ACTUAL_URL = "http://nohost/"
    
    def __init__(self):
        self.RESPONSE = self.response = MockResponse()
        self.SESSION = {}
        self.environ = {}
        self.form = dict(SESSION=dict())

    def __getitem__(self, key):
        return self.form.get(key)


class MockPAS(Acquisition.Implicit):
    
    def __init__(self):
        self.REQUEST = MockRequest()
        
    def updateCredentials(self, request, response, login, new_password):
        pass
        
class FormParser(sgmllib.SGMLParser):
    
    def __init__(self, verbose=0):
        sgmllib.SGMLParser.__init__(self, verbose)
        self.inputs = {}
    
    def parse(self, s):
        self.feed(s)
        self.close()
        
    def start_input(self, attributes):
        name = value = None
        for attr_key, attr_value in attributes:
            if attr_key == 'name':
                name = attr_value
            if attr_key == 'value':
                value = attr_value
        if name is not None and value is not None:
            self.inputs[name] = value


import os
import cgi
from cStringIO import StringIO
import logging
import re

from App.Common import package_home
from AccessControl.SecurityInfo import ClassSecurityInfo
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins \
                import IAuthenticationPlugin, IUserEnumerationPlugin, IExtractionPlugin
from Products.PluggableAuthService.permissions import ManageUsers

from Products.CMFCore.utils import getToolByName

from client import Saml2Client
from saml2.config import Config
from config import sp_config

_browserdir = os.path.join( package_home( globals() ), 'www' )

manage_addSuisseIDPlugin = PageTemplateFile("../www/suisseIDAdd", globals(), 
                __name__="manage_addSuisseIDPlugin")

logger = logging.getLogger("PluggableAuthService")

suisseid_format = re.compile('[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{4}')

attributes = {
    
    # Plain Core Assertion Attributes 
    'Given Names' : 'http://www.ech.ch/xmlns/eCH-0113/1/givenNames',
    'First Name' : 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/givenname',
    'Last Name' : 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/surname',
    'Date of Birth' : 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/dateofbirth',
    # TODO: Date of Birth partially known
    'Place of Birth' : 'http://www.ech.ch/xmlns/eCH-0113/1/placeOfBirth',
    'Origin' : 'http://www.ech.ch/xmlns/eCH-0113/1/origin',
    'Gender' : 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/gender',
    'Nationality' : 'http://www.ech.ch/xmlns/eCH-0113/1/nationality',
    'Identification Number' : 'http://www.ech.ch/xmlns/eCH-0113/1/identificationNumber',
    'Identification Kind' : 'http://www.ech.ch/xmlns/eCH-0113/1/identificationKind',
    'Issuing Country' : 'http://www.ech.ch/xmlns/eCH-0113/1/issuingCountry',
    'Issuing Office' : 'http://www.ech.ch/xmlns/eCH-0113/1/issuingOffice',
    'Identification Issued On' : 'http://www.ech.ch/xmlns/eCH-0113/1/identificationIssuedOn',
    'Identification Valid Until' : 'http://www.ech.ch/xmlns/eCH-0113/1/identificationValidUntil',
    
    # Derived Core Assertion Attributes
    'Age' : 'http://www.ech.ch/xmlns/eCH-0113/1/age',
    'isOver16' : 'http://www.ech.ch/xmlns/eCH-0113/1/isOver16',
    'isOver18' : 'http://www.ech.ch/xmlns/eCH-0113/1/isOver18',
    'age-18-or-over' : 'http://schemas.informationcard.net/@ics/age-18-or-over/2008-11',
    'isSwissCitizen' : 'http://www.ech.ch/xmlns/eCH-0113/1/isSwissCitizen',
}

def addSuisseIDPlugin(self, id, title='', REQUEST=None):
    """Add a suisseID plugin to a Pluggable Authentication Service.
    """
    p=SuisseIDPlugin(id, title)
    self._setObject(p.getId(), p)

    if REQUEST is not None:
        REQUEST["RESPONSE"].redirect("%s/manage_workspace"
                "?manage_tabs_message=suisseID+plugin+added." %
                self.absolute_url())

class SuisseIDPlugin(BasePlugin):
    """suisseID authentication plugin.
    """

    meta_type = "suisseID plugin"
    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._setId(id)
        self.title=title
        self.config = {}
        self._setConfiguration()
        
    def _saml2_config(self):
        if hasattr(self, '_v_cached_config') and self._v_cached_config:
            return self._v_cached_config
        config = Config()
        conf=sp_config.copy()
        metadata_file = self.config['metadata_file']
        if not metadata_file:
            path = os.path.dirname(__file__)
            metadata_file = os.path.join(path, 'metadata.xml')
        conf['metadata']['local'] = [metadata_file]
        config.load(conf)
        config['entityid'] = self.config['portal_url']
        config['service']['sp']['name'] = self.config['portal_name']
        config['service']['sp']['url'] = self.config['portal_url']
        required_attributes = []
        for attribute in self.config['required_attributes'].split('\r\n'):
            name = attributes.get(attribute, None)
            if name:
                required_attributes.append(name)
        optional_attributes = []
        for attribute in self.config['optional_attributes'].split('\r\n'):
            name = attributes.get(attribute, None)
            if name:
                optional_attributes.append(name)
        config['service']['sp']['required_attributes'] = required_attributes
        config['service']['sp']['optional_attributes'] = optional_attributes
        config['service']['sp']['privacy_notice'] = self.config['privacy_notice']
        config['key_file'] = self.config['key_file']
        config['cert_file'] = self.config['cert_file']
        config['xmlsec_binary'] = self.config['xmlsec_binary']
        
        # Get Idps from the metadata
        config['service']['sp']['idp'] = {}
        for location in config['metadata'].locations():
            name = config['metadata'].name(location)
            config['service']['sp']['idp'][name] = location
        
        self._v_cached_config = config
        return self._v_cached_config
        
    def getProviders(self):
        config = self._saml2_config()
        providers = config['service']['sp']['idp'].copy()
        # TODO: Mismatch between actual URL and issuer ID in SAML response
        for name, url in providers.items():
            if url[-1] != '/':
                providers[name] = url + '/'

        return providers
        
        
    # IExtractionPlugin implementation
    def extractCredentials(self, request):
        """This method performs the PAS credential extraction.
        """
        
        creds={}
        # Initiate challenge
        provider_url = request.form.get("__ac_suisseid_provider_url", None)
        if provider_url is not None and provider_url != "":
            config = self._saml2_config()
            scl = Saml2Client(request.environ, config)

            (sid, result) = scl.authenticate(config['entityid'],
                                             provider_url,
                                             config["service"]["sp"]['url'],
                                             config["service"]["sp"]['name'],
                                             log=logger,
                                             required_attributes=config["service"]["sp"]['required_attributes'],
                                             optional_attributes=config["service"]["sp"]['optional_attributes'],
                                             privacy_notice=config['service']['sp']['privacy_notice'])
            
            request.SESSION['suisseid'] = {}
            request.SESSION['suisseid'][sid] = ''
            
            # Compose POST form with onload submit
            form_body = ''.join(result)
            request.response.setHeader("Content-type", "text/html")
            request.response.setHeader("Content-length", str(len(form_body)))
            request.response.setBody(form_body, lock=True)
            request.response.setStatus(200, lock=True)
            return None
        
        # Idp response
        if 'SAMLResponse' in request.form:
      
            post_env = request.environ.copy()
            post_env['QUERY_STRING'] = ''
            
            request.stdin.seek(0)
            post = cgi.FieldStorage(
                fp = StringIO(request.stdin.read()),
                environ = post_env,
                keep_blank_values = True,
            )
            
            config = self._saml2_config()
            scl = Saml2Client(request.environ, config)
            
            session_info = scl.response(post, config['entityid'], request.SESSION.get('suisseid', {}), logger)
            if not session_info:
                return None
            ava = session_info['ava'].copy()
            user_id = ava['__userid']
            del ava['__userid']
            
            creds['suisseid.source'] = 'server'
            creds['suisseid.attributes'] = ava
            creds['login'] = user_id
        
        return creds
            
    # IAuthenticationPlugin implementation
    def authenticateCredentials(self, credentials):
        if not credentials.has_key("suisseid.source"):
            return None
            
        if credentials['suisseid.source'] == 'server':
            identity = credentials.get('login', None)
            if not identity:
                return None
            
            # Use another plugin to store the credentials
            self._getPAS().updateCredentials(self.REQUEST,
                    self.REQUEST.RESPONSE, identity, "")
              
            # That's Plone specific!!!
            if hasattr(self, 'portal_membership'):
                mt = getToolByName(self, 'portal_membership')
                member = mt.getMemberById(credentials['login'])
                attributes = credentials['suisseid.attributes']
                first_name = attributes.get('First Name', [''])[0]
                last_name = attributes.get('Last Name', [''])[0]
                email = attributes.get('Email', [''])[0]
                fullname = ' '.join((first_name, last_name)).strip()
                properties = {}
                if fullname and not member.getProperty('fullname'):
                    properties['fullname'] = fullname
                if email and not member.getProperty('email'):
                    properties['email'] = email
                member.setMemberProperties(properties)
                    
            return (identity, identity)
            
    # IUserEnumerationPlugin implementation
    def enumerateUsers(self, id=None, login=None, exact_match=False,
            sort_by=None, max_results=None, **kw):
            
        if id and login and id!=login:
            return None

        if (id and not exact_match) or kw:
            return None

        key=id and id or login
        
        if suisseid_format.match(key) is None:
            return None
        
        return [ {
                    "id" : key,
                    "login" : key,
                    "pluginid" : self.getId(),
                } ]
     
    #   
    # ZMI configuration tab
    #    
        
    def _setConfiguration(self, 
                          portal_name='', 
                          portal_url='', 
                          required_attributes='',
                          optional_attributes='',
                          privacy_notice='',
                          key_file='', 
                          cert_file='', 
                          xmlsec_binary='/usr/bin/xmlsec1',
                          metadata_file=None):
                          
        self.config['portal_name'] = portal_name
        self.config['portal_url'] = portal_url
        self.config['required_attributes'] = required_attributes
        self.config['optional_attributes'] = optional_attributes
        self.config['privacy_notice'] = privacy_notice
        self.config['key_file'] = key_file
        self.config['cert_file'] = cert_file
        self.config['xmlsec_binary'] = xmlsec_binary
        self.config['metadata_file'] = metadata_file
        
        self._v_cached_config = None
        self._p_changed = 1
        
    security.declareProtected(ManageUsers, 'getConfiguration')
    def getConfiguration(self):
        return self.config
        
    security.declareProtected(ManageUsers, 'changeConfiguration')
    def changeConfiguration(self, portal_name, portal_url, required_attributes,
                            optional_attributes, privacy_notice,
                            key_file, cert_file, xmlsec_binary, metadata_file=None):
                            
        self._setConfiguration(portal_name, portal_url, required_attributes, 
                               optional_attributes, privacy_notice, key_file, 
                               cert_file, xmlsec_binary, metadata_file)
        
    security.declareProtected(ManageUsers, 'manage_editConfiguration')
    def manage_editConfiguration(self, REQUEST=None):
        """Form action for editing configuration.
        """
        if not REQUEST:
            return
        form = REQUEST.form
        portal_name = form.get('portal_name', '').strip()
        portal_url = form.get('portal_url', '').strip()
        required_attributes = form.get('required_attributes', [])
        optional_attributes = form.get('optional_attributes', [])
        privacy_notice = form.get('privacy_notice', '')
        key_file = form.get('key_file', '')
        cert_file = form.get('cert_file', '')
        xmlsec_binary = form.get('xmlsec_binary', '/usr/bin/xmlsec1')
        
        self.changeConfiguration(portal_name, portal_url, required_attributes, optional_attributes,
                                 privacy_notice, key_file, cert_file, xmlsec_binary)
                               
        return REQUEST.RESPONSE.redirect(self.absolute_url() +
                                         '/manage_SuisseIDSettings')

    security.declareProtected(ManageUsers, 'manage_SuisseIDSettings')
    manage_SuisseIDSettings = PageTemplateFile(
        os.path.join(_browserdir, 'editSuisseIDSettings'),
        globals(),
        __name__='manage_SuisseIDSettings')

    manage_options = ({
        'label' : 'suisseID Settings',
        'action' : 'manage_SuisseIDSettings'
    },) + BasePlugin.manage_options
        
classImplements(SuisseIDPlugin, IExtractionPlugin, IAuthenticationPlugin, IUserEnumerationPlugin)



import os
path = os.path.dirname(__file__)

sp_config = {
    
    "entityid" : "",
    "service": {
        "sp":{
            "name" : "",
            "url" : "",
            "required_attributes": [],
            "optional_attributes": [],
            "privacy_notice": "",
            "idp": { },
        }
    },
    "metadata" : {
        "local" : [],
    },
    "debug" : 0,
    "key_file" : "",
    "cert_file" : "",
    "xmlsec_binary" : "/usr/bin/xmlsec1",
    "organization": {
        "name": "",
        "display_name": "",
        "url":"",
    },
    "contact": [{
        "given_name":"",
        "sur_name": "",
        "email_address": "",
        "contact_type": "",
    }]
    
}



import base64
import unittest
import os
import urllib
from StringIO import StringIO
from datetime import datetime, timedelta

path = os.path.dirname(__file__)
xmlsec_binary = '/usr/local/bin/xmlsec1'

class TestSuisseIdExtraction(unittest.TestCase):

    def createPlugin(self):
        from pas.plugins.suisseid.tests.utils import MockPAS
        from pas.plugins.suisseid.plugin import SuisseIDPlugin
        plugin = SuisseIDPlugin("suisseid")
        plugin.changeConfiguration('suisseID Test Portal', 'http://nohost/', '',
                                   '', '', '', '', xmlsec_binary)
        pas = MockPAS()
        return plugin.__of__(pas)
        
    def createIdpResponse(self, authn_request_id='2aaaeb7692471eb4ba00d5546877a7fd'):
        from saml2.saml import Issuer, Assertion
        from saml2.saml import Subject, NameID, SubjectConfirmation, SubjectConfirmationData
        from saml2.saml import Conditions, AudienceRestriction, Audience, OneTimeUse
        from saml2.saml import AuthnStatement, AuthnContext, AuthnContextClassRef
        from saml2.saml import NAMEID_FORMAT_UNSPECIFIED, SUBJECT_CONFIRMATION_METHOD_BEARER
        from saml2.saml import NAMEID_FORMAT_ENTITY
        from saml2.samlp import Response, Status, StatusCode
        from saml2.samlp import STATUS_SUCCESS
        from saml2.utils import make_instance
        from saml2.sigver import pre_signature_part
        from xmldsig import Signature
        
        issue_instant = datetime.utcnow().isoformat() + 'Z'
        not_before = (datetime.utcnow() - timedelta(minutes=5)).isoformat() + 'Z'
        not_on_or_after = (datetime.utcnow() + timedelta(minutes=5)).isoformat() + 'Z'
        issuer = Issuer(format=NAMEID_FORMAT_ENTITY, text='https://idp.swisssign.net/suisseid/eidp')
        signature = make_instance(Signature, pre_signature_part('_ea7f4526-43a3-42d6-a0bc-8f367e95802f'))
        status = Status(status_code=StatusCode(value=STATUS_SUCCESS))
        subject_confirmation_data = SubjectConfirmationData(not_on_or_after=not_on_or_after,
                                                            in_response_to=authn_request_id,
                                                            recipient='http://nohost/')
        subject_confirmation = SubjectConfirmation(method=SUBJECT_CONFIRMATION_METHOD_BEARER,
                                                   subject_confirmation_data=subject_confirmation_data)
        subject = Subject(name_id=NameID(text='1234-1234-1234-1234', format=NAMEID_FORMAT_UNSPECIFIED), 
                          subject_confirmation=subject_confirmation)
        conditions = Conditions(not_before=not_before,
                                not_on_or_after=not_on_or_after,
                                audience_restriction=AudienceRestriction(Audience('http://nohost/')),
                                one_time_use=OneTimeUse())
        authn_context = AuthnContext(authn_context_decl_ref=AuthnContextClassRef('urn:oasis:names:tc:SAML:2.0:ac:classes:SmartcardPKI'))
        authn_statement = AuthnStatement(authn_instant=issue_instant,
                                         authn_context=authn_context)
        assertion_signature = make_instance(Signature, pre_signature_part('_cb8e7dc8-00b3-4655-ad2d-d083cae5168d'))
        assertion = Assertion(id='_cb8e7dc8-00b3-4655-ad2d-d083cae5168d',
                              version='2.0',
                              issue_instant=issue_instant,
                              issuer=issuer,
                              signature=assertion_signature,
                              subject=subject,
                              conditions=conditions,
                              authn_statement=authn_statement,
                             )
        
        response = Response(id='_ea7f4526-43a3-42d6-a0bc-8f367e95802f',
                            in_response_to=authn_request_id,
                            version='2.0',
                            issue_instant=issue_instant,
                            destination='http://nohost/',
                            issuer=issuer,
                            signature=signature,
                            status=status,
                            assertion=assertion,
                           )
        
        return response
        
    def sign_response(self, response):
        from saml2.sigver import sign_statement_using_xmlsec
        response = '%s' % response
        # Sign assertion in the response
        signed_response = sign_statement_using_xmlsec(response, "Response",
                                xmlsec_binary,
                                key_file=os.path.join(path, 'data', 'idp.key'))
        return signed_response
        
    def testNoProviderExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        request = MockRequest()
        result = plugin.extractCredentials(request)
        self.assertEquals(result, {})
        
    def testEmptyProviderExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        request = MockRequest()
        request.form['__ac_suisseid_provider_url'] = ''
        result = plugin.extractCredentials(request)
        self.assertEquals(result, {})
        
    def testAuthnRequestExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest, FormParser
        from saml2.samlp import authn_request_from_string
        plugin = self.createPlugin()
        request = MockRequest()
        request.form['__ac_suisseid_provider_url'] = 'https://idp.swisssign.net/suisseid/eidp/'
        result = plugin.extractCredentials(request)
        # No credentials since that's only the first step
        self.assertEquals(result, None)
        response = request.response
        self.assertEquals(response.status, 200)
        self.assertEquals(response.headers['Content-type'], 'text/html')
        parser = FormParser()
        parser.parse(response.body)
        saml_request = parser.inputs['SAMLRequest']
        decoded_xml = base64.b64decode(saml_request)
        request = authn_request_from_string(decoded_xml)
        self.assertEquals(request.destination, 'https://idp.swisssign.net/suisseid/eidp/')
        self.assertEquals(request.assertion_consumer_service_url, plugin.getConfiguration()['portal_url'])
        
    def testAuthnRequestSignedExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest, FormParser
        from saml2.samlp import authn_request_from_string
        plugin = self.createPlugin()
        sp_pem = os.path.join(path, 'data', 'sp.pem')
        sp_key = os.path.join(path, 'data', 'sp.key')
        plugin.changeConfiguration('suisseID Test Portal', 'http://nohost/', '',
                                   '', '', sp_key, sp_pem, xmlsec_binary)
        request = MockRequest()
        request.form['__ac_suisseid_provider_url'] = 'https://idp.swisssign.net/suisseid/eidp/'
        plugin.extractCredentials(request)
        parser = FormParser()
        parser.parse(request.response.body)
        saml_request = parser.inputs['SAMLRequest']
        decoded_xml = base64.b64decode(saml_request)
        request = authn_request_from_string(decoded_xml)
        self.assertEquals(request.destination, 'https://idp.swisssign.net/suisseid/eidp/')
        # Verify signature
        from saml2.sigver import verify_signature
        verified = verify_signature(xmlsec_binary, decoded_xml, sp_pem, 
                                    node_name='urn:oasis:names:tc:SAML:2.0:protocol:AuthnRequest', 
                                    cert_type='pem')
        self.assertEquals(verified, True)
        
    def testExtendedAuthnRequestExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest, FormParser
        from saml2.samlp import authn_request_from_string
        plugin = self.createPlugin()
        request = MockRequest()
        request.form['__ac_suisseid_provider_url'] = 'https://idp.swisssign.net/suisseid/eidp/'
        # Request three attributes: First Name (required), Last Name (required) and isOver18 (optional)
        plugin.changeConfiguration('suisseID Test Portal', 'http://nohost/', 'First Name\r\nLast Name',
                                   'isOver18', '', '', '', xmlsec_binary)
                                   
        request = MockRequest()
        request.form['__ac_suisseid_provider_url'] = 'https://idp.swisssign.net/suisseid/eidp/'
        plugin.extractCredentials(request)
        parser = FormParser()
        parser.parse(request.response.body)
        saml_request = parser.inputs['SAMLRequest']
        decoded_xml = base64.b64decode(saml_request)
        request = authn_request_from_string(decoded_xml)
        self.assertEquals(request.force_authn, 'true')
        extenions = request.extensions.extension_elements
        self.assertEquals(len(extenions), 3)
        self.assertEquals(extenions[0].attributes['Name'], 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/givenname')
        self.assertEquals(extenions[0].attributes['{http://www.ech.ch/xmlns/eCH-0113/1}required'], 'true')
        
    # Second step of Authentication (SAML response from Idp)
    
    def testResponseExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        plugin.changeConfiguration('', 'http://nohost/', '', '', '', '', '', '/usr/local/bin/xmlsec1',
                                   os.path.join(path, 'data', 'metadata.xml'))
        request = MockRequest()
        # There has to be an outstanding AuthnRequest
        request.SESSION['suisseid'] = { '2aaaeb7692471eb4ba00d5546877a7fd' : '' }
        # Create a SAML2 response
        response = '%s' % self.createIdpResponse('2aaaeb7692471eb4ba00d5546877a7fd')
        signed_response = self.sign_response(response)
        encoded_response = base64.b64encode(signed_response)
        request.form['SAMLResponse'] = encoded_response
        request.environ['REQUEST_METHOD'] = 'POST'
        request.stdin = StringIO(urllib.urlencode({'SAMLResponse' : encoded_response}))
        creds = plugin.extractCredentials(request)
        self.assertEquals(creds['login'], '1234-1234-1234-1234')
        
    def testResponseAuthnFailedExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        plugin.changeConfiguration('', 'http://nohost/', '', '', '', '', '', xmlsec_binary,
                                   os.path.join(path, 'data', 'metadata.xml'))
        request = MockRequest()
        # There has to be an outstanding AuthnRequest
        request.SESSION['suisseid'] = { '2aaaeb7692471eb4ba00d5546877a7fd' : '' }
        # Create a SAML2 response
        response = self.createIdpResponse('2aaaeb7692471eb4ba00d5546877a7fd')
        from saml2.samlp import StatusCode, STATUS_AUTHN_FAILED
        response.status.status_code = StatusCode(value=STATUS_AUTHN_FAILED)
        response = '%s' % response
        signed_response = self.sign_response(response)
        encoded_response = base64.b64encode(signed_response)
        request.form['SAMLResponse'] = encoded_response
        request.environ['REQUEST_METHOD'] = 'POST'
        request.stdin = StringIO(urllib.urlencode({'SAMLResponse' : encoded_response}))
        creds = plugin.extractCredentials(request)
        self.assertEquals(creds, None)
        
    def testResponseManipulatedExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        plugin.changeConfiguration('', 'http://nohost/', '', '', '', '', '', xmlsec_binary,
                                   os.path.join(path, 'data', 'metadata.xml'))
        request = MockRequest()
        # There has to be an outstanding AuthnRequest
        request.SESSION['suisseid'] = { '2aaaeb7692471eb4ba00d5546877a7fd' : '' }
        # Create a SAML2 response
        response = '%s' % self.createIdpResponse('2aaaeb7692471eb4ba00d5546877a7fd')
        signed_response = self.sign_response(response)
        # Response has been manipulated by third party (suisseID number changed).
        signed_response = signed_response.replace('1234-1234-1234-1234', '1234-1234-1234-1235')
        encoded_response = base64.b64encode(signed_response)
        request.form['SAMLResponse'] = encoded_response
        request.environ['REQUEST_METHOD'] = 'POST'
        request.stdin = StringIO(urllib.urlencode({'SAMLResponse' : encoded_response}))
        from saml2.sigver import SignatureError
        self.assertRaises(SignatureError, plugin.extractCredentials, request)
        
    def testResponseNoOutstandingAuthnRequestExtraction(self):
        from pas.plugins.suisseid.tests.utils import MockRequest
        plugin = self.createPlugin()
        plugin.changeConfiguration('', 'http://nohost/', '', '', '', '', '', xmlsec_binary,
                                   os.path.join(path, 'data', 'metadata.xml'))
        request = MockRequest()
        # There are no outstanding AuthnRequesta
        request.SESSION['suisseid'] = {}
        # Create a SAML2 response
        response = '%s' % self.createIdpResponse()
        signed_response = self.sign_response(response)
        encoded_response = base64.b64encode(signed_response)
        request.form['SAMLResponse'] = encoded_response
        request.environ['REQUEST_METHOD'] = 'POST'
        request.stdin = StringIO(urllib.urlencode({'SAMLResponse' : encoded_response}))
        creds = plugin.extractCredentials(request)
        self.assertEquals(creds, None)
        



from AccessControl.Permissions import manage_users as ManageUsers
from Products.PluggableAuthService.PluggableAuthService import registerMultiPlugin

import plugin
registerMultiPlugin(plugin.SuisseIDPlugin.meta_type)

def initialize(context):
    # TODO: Condition on saml2 import
    context.registerClass(plugin.SuisseIDPlugin,
                            permission=ManageUsers,
                            constructors=
                                    (plugin.manage_addSuisseIDPlugin,
                                     plugin.addSuisseIDPlugin),
                            visibility=None,
                            icon="www/suisseid.png")




import unittest

class TestSuisseIdAuthentication(unittest.TestCase):

    def createPlugin(self):
        from pas.plugins.suisseid.tests.utils import MockPAS
        from pas.plugins.suisseid.plugin import SuisseIDPlugin
        plugin = SuisseIDPlugin("suisseid")
        pas = MockPAS()
        return plugin.__of__(pas)
        
    def buildServerResponse(self):
        credentials={}
        credentials['suisseid.source'] = 'server'
        credentials['suisseid.attributes'] = {}
        credentials['login'] = '1234-1234-1234-1234'
        return credentials
        
    def testEmptyAuthentication(self):
        """ Test if we do not invent an identity out of thin air.
        """
        plugin = self.createPlugin()
        creds = plugin.authenticateCredentials({})
        self.assertEqual(creds, None)
        
    def testSuisseIDAuthentication(self):
        """ Test for empty suisseID number
        """
        credentials = self.buildServerResponse()
        plugin = self.createPlugin()
        creds = plugin.authenticateCredentials(credentials)
        self.assertEqual(creds, ('1234-1234-1234-1234', '1234-1234-1234-1234'))


# -*- coding: utf-8 -*-
"""
This module contains the tool of pas.plugins.suisseid
"""
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

version = '0.1b1'

long_description = (
    read('README.txt')
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    read('CONTRIBUTORS.txt')
    )

setup(name='pas.plugins.suisseid',
      version=version,
      description="suisseID PAS plugin for Zope",
      long_description=long_description,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        'Environment :: Web Environment',
        'Framework :: Zope2',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: System :: Systems Administration :: Authentication/Directory'
        ],
      keywords='',
      author='Seantis GmbH',
      author_email='info@seantis.ch',
      url='http://www.seantis.ch',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['pas', 'pas.plugins', ],
      include_package_data=True,
      zip_safe=False,
      install_requires=['setuptools',
                        'httplib2',
                        'python-dateutil',
                        # -*- Extra requirements: -*-
                        ],
      )



from saml2 import SamlBase
from saml2.saml import Attribute, NAMESPACE

class ExtendedAttribute(Attribute):

    c_tag = 'Attribute'
    c_namespace = NAMESPACE
    c_children = Attribute.c_children.copy()
    c_attributes = Attribute.c_attributes.copy()
    c_attributes['eCH-0113:required'] = 'required'
    
    def __init__(self, name=None, name_format=None, friendly_name=None,
                 attribute_value=None, text=None, extension_elements=None,
                 extension_attributes=None, required=False):
                 
        super(ExtendedAttribute, self).__init__(name, name_format, friendly_name,
                                                attribute_value, text, extension_elements,
                                                extension_attributes)
                                                
        self.required = required
        
        
    def _to_element_tree(self):
        element = super(ExtendedAttribute, self)._to_element_tree()
        # We have to define the namespace prefix "eCH-0113" which is used by the "required" attribute
        element.attrib['xmlns:eCH-0113'] = "http://www.ech.ch/xmlns/eCH-0113/1"
        return element
        
        
class PrivacyNotice(SamlBase):
        
    c_tag = 'PrivacyNotice'
    c_namespace = 'http://schemas.xmlsoap.org/ws/2005/05/identity'
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_attributes['Version'] = 'version'
    
    def __init__(self, text=None, extension_elements=None, extension_attributes=None, version='1'):
        super(PrivacyNotice, self).__init__(text, extension_elements, extension_attributes)
        self.version = version
        
    def _to_element_tree(self):
        element = super(PrivacyNotice, self)._to_element_tree()
        # Type element is required (even though it's not in the specification)
        element.attrib['xmlns:xsi'] = "http://www.w3.org/2001/XMLSchema-instance"
        element.attrib['xsi:type'] = "ns1:PrivacyNoticeLocationType"
        return element


import time
import dateutil.parser
import base64
import sys
import os

import xmldsig as ds
import saml2
from saml2 import samlp, saml
from saml2 import VERSION, class_name
from saml2.time_util import instant
from saml2.utils import sid, make_instance
from saml2.sigver import XMLSEC_BINARY, _TEST_
from saml2.sigver import SignatureError, make_temp, cert_from_instance, verify_signature
from saml2.sigver import pre_signature_part, sign_statement_using_xmlsec
from saml2.client import Saml2Client as BaseClient
from saml2.client import for_me

from ech0113 import ExtendedAttribute, PrivacyNotice

FORM_SPEC = """<form method="post" action="%s">
   <input type="hidden" name="SAMLRequest" value="%s" />
   <input type="hidden" name="RelayState" value="%s" />
   <noscript><p>Click 'Continue' to send the request manually.</p><input type="submit" value="Continue" /></noscript>
</form>"""

RESPONSE_NODE = 'urn:oasis:names:tc:SAML:2.0:protocol:Response'

class Saml2Client(BaseClient):
    
    def extended_authn_request(self, query_id, destination, service_url, spentityid, 
                        my_name, vorg="", scoping=None, log=None, sign=False,
                        required_attributes=[], optional_attributes=[], privacy_notice=None):
        """ Creates an authentication request.
        
        :param query_id: The identifier for this request
        :param destination: Where the request should be sent.
        :param service_url: Where the reply should be sent.
        :param spentityid: The entity identifier for this service.
        :param my_name: The name of this service.
        :param vorg: The vitual organization the service belongs to.
        :param scoping: The scope of the request
        :param log: A service to which logs should be written
        :param sign: Whether the request should be signed or not.
        :param required_attributes: Required attributes
        :param optional_attributes: Optional attributes
        """
        prel = {
            "id": query_id,
            "version": VERSION,
            "issue_instant": instant(),
            "destination": destination,
            "assertion_consumer_service_url": service_url,
            "protocol_binding": saml2.BINDING_HTTP_POST,
            "provider_name": my_name,
        }
        
        if scoping:
            prel["scoping"] = scoping
            
        name_id_policy = {
            "allow_create": "true"
        }
        
        name_id_policy["format"] = saml.NAMEID_FORMAT_TRANSIENT
        if vorg:
            try:
                name_id_policy["sp_name_qualifier"] = vorg
                name_id_policy["format"] = saml.NAMEID_FORMAT_PERSISTENT
            except KeyError:
                pass
        
        if sign:
            prel["signature"] = pre_signature_part(prel["id"])

        prel["name_id_policy"] = name_id_policy
        prel["issuer"] = { "text": spentityid }
        
        if log:
            log.info("DICT VERSION: %s" % prel)
            
        request = make_instance(samlp.AuthnRequest, prel)
        
        extensions = []
        for attribute in required_attributes:
            name_format = 'urn:oasis:names:tc:SAML:2.0:attrname-format:uri'
            extensions.append(ExtendedAttribute(name_format=name_format, name=attribute, required='true'))
        for attribute in optional_attributes:
            name_format = 'urn:oasis:names:tc:SAML:2.0:attrname-format:uri'
            extensions.append(saml.Attribute(name_format=name_format, name=attribute))
        
        if privacy_notice:
            extensions.append(PrivacyNotice(text=privacy_notice))
            
        if extensions:
            request.extensions = samlp.Extensions(extension_elements=extensions)
            request.force_authn = 'true'
        
        if sign:
            
            # Add public key to Signature
            fd = open(self.config["cert_file"], 'r')
            pub_key = fd.read()
            fd.close()
            pub_key = pub_key.replace('-----BEGIN CERTIFICATE-----', '')
            pub_key = pub_key.replace('-----END CERTIFICATE-----', '')
            pub_key = pub_key.replace('\n', '')
            x509_certificate = ds.X509Certificate(pub_key)
            x509_data = ds.X509Data(x509_certificate=x509_certificate)
            request.signature.key_info = ds.KeyInfo(x509_data=x509_data)
            
            return sign_statement_using_xmlsec("%s" % request, class_name(request),
                                    self.config["xmlsec_binary"], 
                                    key_file=self.config["key_file"])
                                    
            #return samlp.authn_request_from_string(sreq)
        else:
            return "%s" % request
            
    def authenticate(self, spentityid, location="", service_url="", 
                        my_name="", relay_state="",
                        binding=saml2.BINDING_HTTP_POST, log=None,
                        vorg="", scoping=None,
                        required_attributes=[], optional_attributes=[],
                        privacy_notice=None):
        """ Either verifies an authentication Response or if none is present
        send an authentication request.
        
        :param spentityid: The SP EntityID
        :param binding: How the authentication request should be sent to the 
            IdP
        :param location: Where the IdP is.
        :param service_url: The SP's service URL
        :param my_name: The providers name
        :param relay_state: To where the user should be returned after 
            successfull log in.
        :param binding: Which binding to use for sending the request
        :param log: Where to write log messages
        :param vorg: The entity_id of the virtual organization I'm a member of
        :param scoping: For which IdPs this query are aimed.
            
        :return: AuthnRequest response
        """
        
        if log:
            log.info("spentityid: %s" % spentityid)
            log.info("location: %s" % location)
            log.info("service_url: %s" % service_url)
            log.info("my_name: %s" % my_name)
        session_id = sid()
        
        sign = self.config["key_file"] and self.config["cert_file"]
        
        authen_req = self.extended_authn_request(session_id, location, 
                                service_url, spentityid, my_name, vorg, 
                                scoping, log, sign,
                                required_attributes=required_attributes, 
                                optional_attributes=optional_attributes,
                                privacy_notice=privacy_notice)
        log and log.info("AuthNReq: %s" % authen_req)
        
        if binding == saml2.BINDING_HTTP_POST:
            # No valid ticket; Send a form to the client
            # THIS IS NOT TO BE USED RIGHT NOW
            response = []
            response.append("<html>")
            response.append("<head>")
            response.append("""<title>SAML 2.0 POST</title>""")
            response.append("</head><body>")
            #login_url = location + '?spentityid=' + "lingon.catalogix.se"
            response.append(FORM_SPEC % (location, base64.b64encode(authen_req), relay_state))
            response.append("""<script type="text/javascript">""")
            response.append("     window.onload = function ()")
            response.append(" { document.forms[0].submit(); }")
            response.append("""</script>""")
            response.append("</body>")
            response.append("</html>")
        elif binding == saml2.BINDING_HTTP_REDIRECT:
            raise Exception("HTTP redirect binding type not supported by suisseID" )
        else:
            raise Exception("Unkown binding type: %s" % binding)
        return (session_id, response)
    
    def _verify_condition(self, assertion, requestor, log, lax=False, 
                        slack=0):
        # The Identity Provider MUST include a <saml:Conditions> element
        #print "Conditions",assertion.conditions
        assert assertion.conditions
        condition = assertion.conditions
        log and log.info("condition: %s" % condition)

        try:
            slack = self.config["accept_time_diff"]
        except KeyError:
            slack = 0

        try:
            not_on_or_after = _use_on_or_after(condition, slack)
            _use_before(condition, slack)
        except Exception:
            if not lax:
                raise
            else:
                not_on_or_after = 0
            
        if not for_me(condition, requestor):
            raise Exception("Not for me!!!")
    
        return not_on_or_after

    def verify_response(self, xml_response, requestor, outstanding=None, 
                log=None, decode=True, context="", lax=False):
        """ Verify a response
    
        :param xml_response: The response as a XML string
        :param requestor: The hostname of the machine
        :param outstanding: A collection of outstanding authentication requests
        :param log: Where logging information should be sent
        :param decode: There for testing purposes
        :param lax: Accept things you normally shouldn't
        :return: A 2-tuple consisting of an identity description and the 
            real relay-state
        """
    
        if not outstanding:
            outstanding = {}
    
        if decode:
            decoded_xml = base64.b64decode(xml_response)
        else:
            decoded_xml = xml_response
    
        # own copy
        xmlstr = decoded_xml[:]
        log and log.info("verify correct signature")
        # IdP/CAS must sign assertion, thus must=True
        response = correctly_signed_response(decoded_xml, 
                        self.config["xmlsec_binary"], log=log,
                        metadata=self.config['metadata'])
        if not response:
            if log:
                log.error("Response was not correctly signed")
                log.info(decoded_xml)
            return None
        else:
            log and log.info("Response was correctly signed")
        
        log and log.info("response: %s" % (response,))
        try:
            session_info = self.do_response(response, 
                                            requestor, 
                                            outstanding=outstanding, 
                                            xmlstr=xmlstr, 
                                            log=log, 
                                            context=context,
                                            lax=lax)
            session_info["issuer"] = response.issuer.text
            session_info["session_id"] = response.in_response_to
        except AttributeError, exc:
            if log:
                log.error("AttributeError: %s" % (exc,))
            else:
                print >> sys.stderr, "AttributeError: %s" % (exc,)
            return None
        except Exception, exc:
            if log:
                log.error("Exception: %s" % (exc,))
            else:
                print >> sys.stderr, "Exception: %s" % (exc,)
            return None
                                
        session_info["ava"]["__userid"] = session_info["name_id"]
        return session_info
        
        
def correctly_signed_response(decoded_xml,
        xmlsec_binary=XMLSEC_BINARY, metadata=None, log=None, must=False):
    """ Check if a instance is correctly signed, if we have metadata for
    the IdP that sent the info use that, if not use the key that are in 
    the message if any.

    :param decode_xml: The SAML message as a XML string
    :param xmlsec_binary: Where the xmlsec1 binary can be found on this
        system.
    :param metadata: Metadata information
    :return: None if the signature can not be verified otherwise an instance
    """

    print "-- correctly_signed_response --"
    response = samlp.response_from_string(decoded_xml)

    if not xmlsec_binary:
        xmlsec_binary = XMLSEC_BINARY

    # Try to find the signing cert in the assertion
    for assertion in response.assertion:
        if not assertion.signature:
            if _TEST_:
                log and log.info("unsigned")
            if must:
                raise SignatureError("Signature missing")
            continue
        else:
            if _TEST_:
                log and log.info("signed")
    
        issuer = assertion.issuer.text.strip()
        if _TEST_:
            print "issuer: %s" % issuer
        if metadata:
            certs = metadata.certs(issuer)
        else:
            certs = []

        if _TEST_:
            print "metadata certs: %s" % certs

        if not certs:
            certs = [make_temp("%s" % cert, ".der") \
                        for cert in cert_from_instance(assertion)]
        if not certs:
            raise SignatureError("Missing certificate")

        verified = False
        for _, der_file in certs:
            if verify_signature(xmlsec_binary, decoded_xml, der_file, node_name=RESPONSE_NODE):
                verified = True
                break
                
        if not verified:
            raise SignatureError("Could not verify")

    return response
   

def _use_on_or_after(condition, slack):
    now = time.mktime(time.gmtime())
    not_on_or_after = dateutil.parser.parse(condition.not_on_or_after)
    not_on_or_after = time.mktime(not_on_or_after.utctimetuple())
    if not_on_or_after < now + slack:
        # To old ignore
        raise Exception("To old can't use it (%s < %s)" % ( not_on_or_after,
                        (now + slack)))
    return not_on_or_after
    
def _use_before(condition, slack):
    not_before = dateutil.parser.parse(condition.not_before)
    not_before = time.mktime(not_before.utctimetuple())
    now = time.mktime(time.gmtime())
        
    if not_before > now + slack:
        # Can't use it yet
        raise Exception("Can't use it yet %s > %s" % (not_before, now)) 